

SubgroupAutomorphismGroup := function(g) 
  local gens,augens,e,i,eord,val,j,hom,t,ok,g2;
  gens := GeneratorsOfGroup(g);
  augens := List(gens,x -> ConjugatorAutomorphism(g,x));
  e := Elements(g);
  for t in [1..5*Size(gens)*Size(gens)] do
  g2 := [];
  eord := List(gens,x -> Filtered(e,y -> Order(y)=Order(x)));
  ok := true;
  for i in [1..Length(gens)] do
    val := eord[i];
    for j in [1..i-1] do
      val := Filtered(val,x -> Order(x*g2[j])=Order(gens[i]*gens[j]));
    od;
    if Size(val) = 0 then
      ok := false;
      break;
    fi;
    Append(g2,[Random(val)]);
  od;
  if (ok) then
    hom := GroupHomomorphismByImages(g,g,gens,g2);

    if not (hom = fail) then
      if IsBijective(hom) then
        Append(augens,[hom]);
      fi;
    fi;
  fi;
  od;
  return augens;
end;





RankList := function(ap) 
  local rks,mx;
  if Size(ap) = 0 then return []; fi;
  rks := List(ap,Size);
  mx := Maximum(rks);
  return List([1..mx],i -> Size(Filtered(rks,j -> j=i)));
end;



IsPolytope := function(gen) 
  local tt,n,st,gg;
# first, check that all generators are involutions 
  st := Set(gen,Order);
  if not st = [2] then
     return false;
  fi;
# first, check that the ggi has a string diagram 
  n := Length(gen);
  if (n > 2) then
    tt := Tuples([1..n],2);
    tt := Filtered(tt,t -> t[2] > t[1]+1);
    st := Set(tt,t -> Order(gen[t[1]]*gen[t[2]]));
    if not st = [2] then
      return false;
    fi;
  fi;
# lastly, check the intersection property
# all I and J
  tt := Tuples(Combinations([1..n]),2);
# eliminate duplicated tests
  tt := Filtered(tt,t -> t[1] < t[2]);
  tt := List(tt,t -> [t[1],t[2],Intersection(t[1],t[2]),Union(t[1],t[2])]);
# no need to check the case when I or J is a subset of the other
  tt := Filtered(tt, t -> not t[1]=t[3]);
  tt := Filtered(tt, t -> not t[2]=t[3]);
# nor the case where I and J don't span an interval
  tt := Filtered(tt, t -> t[4] = [Minimum(t[4])..Maximum(t[4])]);
  tt := List(tt,t -> [t[1],t[2],t[3]]);
  gg := List(tt,t -> List(t,j -> List(j,i -> gen[i])));
  gg := List(gg,g -> [g[1],g[2],Concatenation(g[3],[()])]);
  gg := List(gg,g -> List(g,Group));
  return ForAll(gg,g -> Size(Intersection(g[1],g[2]))=Size(g[3]));
end;

Mix := function(ga,gb)
  local A,B,pA,pB,gc,mv;
  A := Group(ga);
  B := Group(gb);
  pA := Maximum(MovedPoints(A));
  pB := Maximum(MovedPoints(B));
  mv := List([1..pB],i -> (i,i+pA));
  mv := Product(mv)^-1;
  gc := List([1..Size(gb)],i -> gb[i]^(mv)*ga[i]);
  return gc;
end;

IsIsom := function(p1,p2) 
  local hom1;
  hom1 := GroupHomomorphismByImages(Group(p1),Group(p2),p1,p2);
  if hom1 = fail then return false; fi;
  return IsBijective(hom1);
end;

  
Dual := function(ga)
  return Reversed(ga);
end;

IsSelfDual := function(pp) 
  return IsIsom(pp,Dual(pp));
end;


ConjugacyClassesInvolutions := function(grp)
   local ss,inv,cr,cc;
   ss := SylowSubgroup(grp,2);
   inv := Filtered(Elements(ss),x -> Order(x) = 2);
   cr := [];
   while (Size(inv) > 0) do
      Append(cr,[inv[1]]);
      cc := ConjugacyClass(grp,inv[1]);
      inv := Filtered(inv, x -> not x in cc);
   od;
   return cr;
end;


ConjugacyClassesElementsP := function(grp,q)
   local ss,inv,cr,p;
   if not Size(Set(Factors(q))) = 1 then return fail; fi;
   p := Factors(q)[1];
   ss := SylowSubgroup(grp,p);
   inv := Filtered(Elements(ss),x -> Order(x) = q);
   cr := [];
   while (Size(inv) > 0) do
      Append(cr,[inv[1]]);
      inv := Filtered(inv, x -> not IsConjugate(grp,x,inv[1]));
   od;
   return cr;
end;



AutomorphismClassRepresentativeInvolutions := function ( G )

local  ccl, rep, A, autgen, outer, pos, Out, orb, i;

  ccl    := ConjugacyClassesInvolutions(G);
  if Length( Set ( List( ccl, x -> Size(ConjugacyClass(G,x)) ) ) ) = Length( ccl )
  then return ccl; fi;
  rep    := ccl;
  A      := AutomorphismGroup( G );
  autgen := GeneratorsOfGroup( A );
  outer  := Filtered( autgen, a -> not IsInnerAutomorphism( a ) );
  Out := Group(()); pos := [ ];
  for i in [ 1 .. Length( outer ) ] do
    pos[i] := List([ 1 .. Length( ccl ) ],
                   j ->
Position(List(rep,g->IsConjugate(G,g,Image(outer[i],
                                                                  
rep[j]))),
                                 true));
    Out := ClosureGroup( Out, SortingPerm( pos[ i ] ) );
  od;
  return List( Orbits( Out, [ 1 .. Length( ccl ) ] ),
               orb -> Representative( ccl[ orb[ 1 ] ] ) );
end;





AutomorphismClassesInvolutionsOld := function(grp)
   local ss,inv,cr,au;
   au := AutomorphismGroup(grp);
   ss := SylowSubgroup(grp,2);
   inv := Filtered(Elements(ss),x -> Order(x) = 2);
#   cr := [];
#   while (Size(inv) > 0) do
#      Append(cr,[inv[1]]);
#      inv := Filtered(inv,x -> not x in Orbit(au,inv[1]));
#   od;
   cr := List(Orbits(au,inv),Representative);
   return cr;
end;


AutomorphismClassesInvolutions := function(grp)
#   if (Size(AutomorphismGroup(grp))/Size(grp)) < 10 and Size(grp)>10000 then
#      return AutomorphismClassRepresentativeInvolutions(grp);
#   else 
      return AutomorphismClassesInvolutionsOld(grp);
#   fi;
end;



#AutomorphismClassesInvolutions2 := function(grp)
#   local ss,inv,cr,au;
#   au := AutomorphismGroup(grp);;
#   ss := SylowSubgroup(grp,2);;
#   inv := Filtered(Elements(ss),x -> Order(x) = 2);;
#   cr := [];
#   while (Size(inv) > 0) do
#      Append(cr,[inv[1]]);
#      inv := Filtered(inv,x -> not inv[1] in x^au);
#   od;
#   cr := List(Orbits(au,inv),Representative);
#   return cr;
#end;

AutomorphismClassesElementsP := function(grp,p)
   local ss,inv,cr,au;
   au := AutomorphismGroup(grp);
   ss := SylowSubgroup(grp,p);
   inv := Filtered(Elements(ss),x -> Order(x) = p);
#   cr := [];
#   while (Size(inv) > 0) do
#      Append(cr,[inv[1]]);
#      inv := Filtered(inv,x -> not x in Orbit(au,inv[1]));
#   od;
   cr := List(Orbits(au,inv),Representative);
   return cr;
end;



AllInvolutions := function(grp) 
   local orb,inv;
   inv := ConjugacyClassesInvolutions(grp);
   orb := List(inv,x -> Elements(ConjugacyClass(grp,x)));
   return Union(orb);
end;

AllElementsP := function(grp,p) 
   local orb,ss,inv;
   ss := SylowSubgroup(grp,p);
   inv := Filtered(Elements(ss),x -> Order(x)=p);
   orb := Orbits(grp,inv);
   return Union(orb);
end;


PartialPolytopes := function(grp,n) 
   local cr,c,ccr,ppp,pp,p,pg,stb,els,fels,i,j,k,ok,h0,hn,h0n,rtn;
   if (n < 2) then
      cr := ConjugacyClassesInvolutions(grp);
      return List(cr, x -> [x]);
   fi;
# get partial polytopes of length n-1, ie, potential facets
   pp := PartialPolytopes(grp,n-1);
   els := AllInvolutions(grp);
   ppp := [];
   for p in pp do
     pg := Group(p);
# eliminate partial polytopes (facets) which are full polytopes
     if (Size(pg) < Size(grp)) then  
# stabiliser under g of the partial polytope
#       stb := Stabilizer(AutomorphismGroup(grp),p,OnTuples);
       stb := Stabilizer(grp,p,OnTuples);
       fels := Filtered(els,x -> not x in pg);
# find representative involutions for the orbits of stb
       cr := [];
       while (Size(fels) > 0) do
         Append(cr,[fels[1]]);
#         fels := Filtered(fels, x -> not x in Orbit(stb,fels[1]));
         fels := Filtered(fels, x -> not IsConjugate(stb,x,fels[1]));
       od;
# find those which give string diagrams
       ccr := [];
       for c in cr do
         ok := true;
         for i in [1..n-2] do
           if (Order(c*p[i]) > 2) then
              ok := false;
              break;
           fi;
         od;
         if Order(c*p[n-1]) = 2 then
           ok := false;
         fi;
         if ok then
           Append(ccr,[c]);
         fi;
       od;
# found some.
       Append(ppp,List(ccr,x -> Concatenation(p,[x])));
     fi;     
   od;
# check intersection property
   rtn := [];
   ok := true;
   for pp in ppp do
     for i in [1..n-2] do
       for j in [i+1..n-1] do
         for k in [j-1..n-1] do
           if ok then
             hn := Group(List([i..k], z -> pp[z]));
             h0 := Group(List([j..n], z -> pp[z]));
             h0n := Group([()]);
             if (k >= j) then
               h0n := Group(List([j..k], z -> pp[z]));
             fi;
             if not (Size(h0n) = Size(Intersection(h0,hn))) then
               ok := false;
             fi;
           fi;
         od;
       od;
     od;
     if (ok) then 
       Append(rtn,[pp]);
     fi;
   od;
   return rtn;
end;
 
PolytopeType := function(p) 
  local gg;
  gg := List([1..Length(p)-1], i -> p[i]*p[i+1]);
  return List(gg,Order);
end;
    
Polytopes := function(grp,n) 
  return Filtered(PartialPolytopes(grp,n),p -> Size(Group(p)) = Size(grp));
end;


AllPolytopes1 := function(grp,n,lll)
   local autG,ap,fp,cr,c,ccr,ppp,stpp,stp,pp,p,pg,stb,els,fels,i,j,k,ok,h0,hn,h0n,rtn,rtn2,iii;
   autG := AutomorphismGroup(grp);
   if (n < 2) then
      cr := AutomorphismClassesInvolutions(grp);
      return [List(cr, x -> [[x],Stabilizer(autG,x)]),[]];
   fi;
# get partial polytopes of length n-1, ie, potential facets
   if Size(lll) > 0 then
      els := lll;
   else
      els := AllInvolutions(grp);   
   fi;
   ap := AllPolytopes1(grp,n-1,els);
   fp := ap[2];
   stpp := ap[1];
   Print("## Now starting n = ",n,". In rank < n found ",Size(ap[2])," polytopes and ");
   Print(Size(ap[1])," potential facets\n");
   ppp := [];
   iii := 1;
   rtn2 := fp;
   for stp in stpp do
     Print("## Checking ",iii,", of ",Size(ap[1]));
     Print(". ",RankList(rtn2)," polys so far.\n");
     iii := iii+1;
     p := stp[1];
     stb := stp[2];
     pg := Group(p);
# eliminate partial polytopes (facets) which are full polytopes
#     if (Size(pg) < Size(grp)) then  
# stabiliser under g of the partial polytope
#       stb := Stabilizer(AutomorphismGroup(grp),p,OnTuples);
#       stb := Stabilizer(grp,p,OnTuples);    # could I save time here by storing stb with p ?
       fels := Filtered(els,x -> not x in pg);  # and here?? A fraction perhaps...
# find representative involutions for the orbits of stb
#       cr := [];
#       while (Size(fels) > 0) do
#         Append(cr,[fels[1]]);
#         fels := Filtered(fels, x -> not x in Orbit(stb,fels[1]));
#         fels := Filtered(fels, x -> not IsConjugate(stb,x,fels[1]));    
#                     # is there a gap function for what I am attempting here??
#       od;
       cr := List(Orbits(stb,fels),Representative);
# find those which give string diagrams
       ccr := [];
       for c in cr do
         ok := true;
         for i in [1..n-2] do
           if (Order(c*p[i]) > 2) then    # could invalid elements be stored with each involution ?
              ok := false;		# probably too much memory.
              break;
           fi;
         od;
         if Order(c*p[n-1]) = 2 then
           ok := false;
         fi;
# It's worth checking intersection property here. Backup first!!
         pp := Concatenation(p,[c]);
        if ok then
         for i in [1..n-2] do
           for j in [i+1..n-1] do
             for k in [j-1..n-1] do
               if ok then
                 hn := Group(List([i..k], z -> pp[z]));
                 h0 := Group(List([j..n], z -> pp[z]));
                 h0n := Group([()]);
                 if (k >= j) then
                   h0n := Group(List([j..k], z -> pp[z]));
                 fi;
                 if not (Size(h0n) = Size(Intersection(h0,hn))) then
                   ok := false;
                 fi;
               fi;
             od;
           od;
         od;
        fi;
         if ok then
           if (Size(Group(pp)) = Size(grp)) then
              Append(rtn2,[pp]);
           else 
              Append(ccr,[c]);
           fi;
         fi;
       od;
# found some.
       Append(ppp,List(ccr,x -> [Concatenation(p,[x]),Stabilizer(stb,x)]));
#     fi;     
   od;
   rtn := [ppp,rtn2];
   return rtn;
end; 

AllPolytopesNoAut1 := function(grp,n,lll)
   local ap,fp,cr,c,ccr,ppp,stpp,stp,pp,p,pg,stb,els,fels,i,j,k,ok,h0,hn,h0n,rtn,rtn2,iii;
   if (n < 2) then
      cr := ConjugacyClassesInvolutions(grp);
      return [List(cr, x -> [[x],Stabilizer(grp,x)]),[]];
   fi;
# get partial polytopes of length n-1, ie, potential facets
   if Size(lll) > 0 then
      els := lll;
   else
      els := AllInvolutions(grp);   
   fi;
   ap := AllPolytopesNoAut1(grp,n-1,els);
   fp := ap[2];
   stpp := ap[1];
   Print("## Now starting n = ",n,". In rank < n found ",Size(ap[2])," polytopes and ");
   Print(Size(ap[1])," potential facets\n");
   ppp := [];
   iii := 1;
   rtn2 := fp;
   for stp in stpp do
     Print("## Checking ",iii,", of ",Size(ap[1]),"\n");
     iii := iii+1;
     p := stp[1];
     stb := stp[2];
     pg := Group(p);
# eliminate partial polytopes (facets) which are full polytopes
#     if (Size(pg) < Size(grp)) then  
# stabiliser under g of the partial polytope
#       stb := Stabilizer(AutomorphismGroup(grp),p,OnTuples);
#       stb := Stabilizer(grp,p,OnTuples);    # could I save time here by storing stb with p ?
       fels := Filtered(els,x -> not x in pg);  # and here?? A fraction perhaps...
# find representative involutions for the orbits of stb
#       cr := [];
#       while (Size(fels) > 0) do
#         Append(cr,[fels[1]]);
#         fels := Filtered(fels, x -> not x in Orbit(stb,fels[1]));
#         fels := Filtered(fels, x -> not IsConjugate(stb,x,fels[1]));    
#                     # is there a gap function for what I am attempting here??
#       od;
       cr := List(Orbits(stb,fels),Representative);
# find those which give string diagrams
       ccr := [];
       for c in cr do
         ok := true;
         for i in [1..n-2] do
           if (Order(c*p[i]) > 2) then    # could invalid elements be stored with each involution ?
              ok := false;		# probably too much memory.
              break;
           fi;
         od;
         if Order(c*p[n-1]) = 2 then
           ok := false;
         fi;
# It's worth checking intersection property here. Backup first!!
         pp := Concatenation(p,[c]);
        if ok then
         for i in [1..n-2] do
           for j in [i+1..n-1] do
             for k in [j-1..n-1] do
               if ok then
                 hn := Group(List([i..k], z -> pp[z]));
                 h0 := Group(List([j..n], z -> pp[z]));
                 h0n := Group([()]);
                 if (k >= j) then
                   h0n := Group(List([j..k], z -> pp[z]));
                 fi;
                 if not (Size(h0n) = Size(Intersection(h0,hn))) then
                   ok := false;
                 fi;
               fi;
             od;
           od;
         od;
        fi;
         if ok then
           if (Size(Group(pp)) = Size(grp)) then
              Append(rtn2,[pp]);
           else 
              Append(ccr,[c]);
           fi;
         fi;
       od;
# found some.
       Append(ppp,List(ccr,x -> [Concatenation(p,[x]),Stabilizer(stb,x)]));
#     fi;     
   od;
   rtn := [ppp,rtn2];
   return rtn;
end; 

OnePolytopeNoAut1 := function(grp,n,lll)
   local ap,fp,cr,c,ccr,ppp,stpp,stp,pp,p,pg,stb,els,fels,i,j,k,ok,h0,hn,h0n,rtn,rtn2,iii;
   if (n < 2) then
      cr := ConjugacyClassesInvolutions(grp);
      return [List(cr, x -> [[x],Stabilizer(grp,x)]),[]];
   fi;
# get partial polytopes of length n-1, ie, potential facets
   if Size(lll) > 0 then
      els := lll;
   else
      els := AllInvolutions(grp);   
   fi;
   ap := OnePolytopeNoAut1(grp,n-1,els);
   if Size(ap[2]) > 0 then
     return ap;
   fi;
   fp := ap[2];
   stpp := ap[1];
#   Print("## Now starting n = ",n,". In rank < n found ",Size(ap[2])," polytopes and ");
   Print(Size(ap[1])," potential facets\n");
   ppp := [];
   iii := 1;
   rtn2 := fp;
   for stp in stpp do
#     Print("## Checking ",iii,", of ",Size(ap[1]),"\n");
     iii := iii+1;
     p := stp[1];
     stb := stp[2];
     pg := Group(p);
# eliminate partial polytopes (facets) which are full polytopes
#     if (Size(pg) < Size(grp)) then  
# stabiliser under g of the partial polytope
#       stb := Stabilizer(AutomorphismGroup(grp),p,OnTuples);
#       stb := Stabilizer(grp,p,OnTuples);    # could I save time here by storing stb with p ?
       fels := Filtered(els,x -> not x in pg);  # and here?? A fraction perhaps...
# find representative involutions for the orbits of stb
#       cr := [];
#       while (Size(fels) > 0) do
#         Append(cr,[fels[1]]);
#         fels := Filtered(fels, x -> not x in Orbit(stb,fels[1]));
#         fels := Filtered(fels, x -> not IsConjugate(stb,x,fels[1]));    
#                     # is there a gap function for what I am attempting here??
#       od;
       cr := List(Orbits(stb,fels),Representative);
# find those which give string diagrams
       ccr := [];
       for c in cr do
         ok := true;
         for i in [1..n-2] do
           if (Order(c*p[i]) > 2) then    # could invalid elements be stored with each involution ?
              ok := false;		# probably too much memory.
              break;
           fi;
         od;
         if Order(c*p[n-1]) = 2 then
           ok := false;
         fi;
# It's worth checking intersection property here. Backup first!!
         pp := Concatenation(p,[c]);
        if ok then
         for i in [1..n-2] do
           for j in [i+1..n-1] do
             for k in [j-1..n-1] do
               if ok then
                 hn := Group(List([i..k], z -> pp[z]));
                 h0 := Group(List([j..n], z -> pp[z]));
                 h0n := Group([()]);
                 if (k >= j) then
                   h0n := Group(List([j..k], z -> pp[z]));
                 fi;
                 if not (Size(h0n) = Size(Intersection(h0,hn))) then
                   ok := false;
                 fi;
               fi;
             od;
           od;
         od;
        fi;
         if ok then
           if (Size(Group(pp)) = Size(grp)) then
              Append(rtn2,[pp]);
              return [[],rtn2];
           else 
              Append(ccr,[c]);
           fi;
         fi;
       od;
# found some.
       Append(ppp,List(ccr,x -> [Concatenation(p,[x]),Stabilizer(stb,x)]));
#     fi;     
   od;
   rtn := [ppp,rtn2];
   return rtn;
end; 



BAlt := function(q) 
  local L,n,p,g;
  L := AlternatingGroup(q); 
  n := Size(MovedPoints(L));
  p := Product(List([1..n], i -> (i,i+n)));
  g := List(GeneratorsOfGroup(L));
  Append(g,List(g,h -> h^p));
  Append(g,[p]);
  return Group(g);
end;

BSym := function(q) 
  local L,n,p,g;
  L := SymmetricGroup(q); 
  n := Size(MovedPoints(L));
  p := Product(List([1..n], i -> (i,i+n)));
  g := List(GeneratorsOfGroup(L));
  Append(g,List(g,h -> h^p));
  Append(g,[p]);
  return Group(g);
end;
             


IsGeneratedByInvolutions := function(G)
    local c, U, i;
    c := ConjugacyClasses(G);
    i := Filtered( c, x -> Order(Representative(x)) = 2 );
    U := Subgroup( G, Flat( List( c, Elements ) ) );
    return U = G;
end;
   
AllPolytopes := function(grp)
  local sz,n,autG,si,lll,sy,inv,cc,cr;
  sz := Size(grp);
  Print("## Group of size ",sz,".\n");
  Print("## Preliminary Checking\n");
  if (sz = 2) then 
    Print("## Group has order 2.\n");
    return [[grp.1]]; 
  fi;
  if IsOddInt(sz) then 
    Print("## Group has odd order.\n");
    return []; 
  fi;
#  if not IsGeneratedByInvolutions(grp) then 
#     Print("Group is not a ggi\n");
#     return [];
#  fi;
  sy := SylowSubgroup(grp,2);
  
  if (Size(Core(grp,sy))=Size(sy)) then
     Print("# Sylow Subgrp is normal\n");
     return [];
  fi;
   inv := Filtered(Elements(sy),x -> Order(x) = 2);
   cr := [];
   lll := [];
   while (Size(inv) > 0) do
      Append(cr,[inv[1]]);
      cc := ConjugacyClass(grp,inv[1]);
      Append(lll,Elements(cc));
      inv := Filtered(inv, x -> not x in cc);
   od;
#   if Size(lll) = Size(inv) then
#     Print("# Sylow Subgrp is normal\n");
#     return [];
#  fi;
     
#  lll := AllInvolutions(grp);
  si := Size(Subgroup(grp,lll));
#  if not 2 in Factors(Index(grp,Center(grp))) then 
#    Print("## center has odd index\n");
#    return []; fi;
  if not (si = Size(grp)) then
    Print("## The ",Size(lll)," involutions generate a proper subgroup of size ",si,"\n");
#    Print("## The involutions generate a proper subgroup.\n");
    return [];
  fi;
  if IsPGroup(grp) then
    if IsFinite(grp) then
      if CanEasilyComputePcgs(grp) then
        Print("## Faster Automorphism method being used.\n");
      fi;
    fi;
  fi;
  Print("## Computing Aut(grp)\n");
  SetInfoLevel(InfoAutGrp,1);
  autG := AutomorphismGroup(grp); 
  SetInfoLevel(InfoAutGrp,0);
#  Print("## Automorphism Group of size ",Size(autG),"\n");
  sz := sz/24;
  n := 3; 
  while (sz > 1) do
    n := n+1;
    sz := sz/3;
  od;
#  return List(AllPolytopes1(grp,n,[])[2], x -> x[1]);
#  return AllPolytopes1(grp,n,[])[2];
  return AllPolytopes1(grp,n,lll)[2];
end;

AllPolytopesNoAut := function(grp)
  local sz,n,autG,si,lll,sy,inv,cc,cr;
  sz := Size(grp);
  Print("## Group of size ",sz,".\n");
  Print("## Preliminary Checking\n");
  if (sz = 2) then return [[grp.1]]; fi;
  if IsOddInt(sz) then return []; fi;
#  if not IsGeneratedByInvolutions(grp) then 
#     Print("Group is not a ggi\n");
#     return [];
#  fi;
  sy := SylowSubgroup(grp,2);
  
  if (Size(Core(grp,sy))=Size(sy)) then
     Print("# Sylow Subgrp is normal\n");
     return [];
  fi;
   inv := Filtered(Elements(sy),x -> Order(x) = 2);
   cr := [];
   lll := [];
   while (Size(inv) > 0) do
      Append(cr,[inv[1]]);
      cc := ConjugacyClass(grp,inv[1]);
      Append(lll,Elements(cc));
      inv := Filtered(inv, x -> not x in cc);
   od;
#   if Size(lll) = Size(inv) then
#     Print("# Sylow Subgrp is normal\n");
#     return [];
#  fi;
     
#  lll := AllInvolutions(grp);
  si := Size(Subgroup(grp,lll));
#  if not 2 in Factors(Index(grp,Center(grp))) then 
#    Print("## center has odd index\n");
#    return []; fi;
  if not (si = Size(grp)) then
    Print("## The ",Size(lll)," involutions generate a proper subgroup of size ",si,"\n");
#    Print("## The involutions generate a proper subgroup.\n");
    return [];
  fi;
  sz := sz/24;
  n := 3; 
  while (sz > 1) do
    n := n+1;
    sz := sz/3;
  od;
#  return List(AllPolytopes1(grp,n,[])[2], x -> x[1]);
  return AllPolytopesNoAut1(grp,n,[])[2];
end;


OnePolytopeNoAut := function(grp)
  local sz,n,autG,si,lll,sy,inv,cc,cr;
  sz := Size(grp);
  Print("## Group of size ",sz,".\n");
  Print("## Preliminary Checking\n");
  if (sz = 2) then return [[grp.1]]; fi;
  if IsOddInt(sz) then return []; fi;
#  if not IsGeneratedByInvolutions(grp) then 
#     Print("Group is not a ggi\n");
#     return [];
#  fi;
  sy := SylowSubgroup(grp,2);
  
  if (Size(Core(grp,sy))=Size(sy)) then
     Print("# Sylow Subgrp is normal\n");
     return [];
  fi;
   inv := Filtered(Elements(sy),x -> Order(x) = 2);
   cr := [];
   lll := [];
   while (Size(inv) > 0) do
      Append(cr,[inv[1]]);
      cc := ConjugacyClass(grp,inv[1]);
      Append(lll,Elements(cc));
      inv := Filtered(inv, x -> not x in cc);
   od;
#   if Size(lll) = Size(inv) then
#     Print("# Sylow Subgrp is normal\n");
#     return [];
#  fi;
     
#  lll := AllInvolutions(grp);
  si := Size(Subgroup(grp,lll));
#  if not 2 in Factors(Index(grp,Center(grp))) then 
#    Print("## center has odd index\n");
#    return []; fi;
  if not (si = Size(grp)) then
    Print("## The ",Size(lll)," involutions generate a proper subgroup of size ",si,"\n");
#    Print("## The involutions generate a proper subgroup.\n");
    return [];
  fi;
  sz := sz/24;
  n := 3; 
  while (sz > 1) do
    n := n+1;
    sz := sz/3;
  od;
#  return List(AllPolytopes1(grp,n,[])[2], x -> x[1]);
  return OnePolytopeNoAut1(grp,n,[])[2];
end;

PolytopalGroupList := function(n)
  local i,ap,lis,m;
  lis := [];
  m := NrSmallGroups(n);
  i := 1;
  while i <= m do
    Print("#*********  ",i," of ",m,"\n"); 
    if Size(OnePolytopeNoAut(SmallGroup(n,i))) > 0 then
       Append(lis,[i]);
    fi;
    i := i+1;
  od;
  return lis;
end;

    


SomePolytopes1 := function(grp,n,lll,type)
   local autG,ap,fp,cr,c,ccr,ppp,stpp,stp,pp,p,pg,stb,els,fels,i,j,k,ok,h0,hn,h0n,rtn,iii;
   autG := AutomorphismGroup(grp);
   if (n < 2) then
      cr := AutomorphismClassesInvolutions(grp);
      return [List(cr, x -> [[x],Stabilizer(autG,x)]),[]];
   fi;
# get partial polytopes of length n-1, ie, potential facets
   if Size(lll) > 0 then
      els := lll;
   else
      els := AllInvolutions(grp);   # could save time here by passing this result down
   fi;
   ap := SomePolytopes1(grp,n-1,els,type);
   fp := ap[2];
   stpp := ap[1];
   Print("## Now starting n = ",n,". In rank < n found ",Size(ap[2])," polytopes and ");
   Print(Size(ap[1])," potential facets\n");
   ppp := [];
   iii := 1;
   for stp in stpp do
     Print("## Checking ",iii,", of ",Size(ap[1]),"\n");
     iii := iii+1;
     p := stp[1];
     stb := stp[2];
     pg := Group(p);
# eliminate partial polytopes (facets) which are full polytopes
#     if (Size(pg) < Size(grp)) then  
# stabiliser under g of the partial polytope
#       stb := Stabilizer(AutomorphismGroup(grp),p,OnTuples);
#       stb := Stabilizer(grp,p,OnTuples);    # could I save time here by storing stb with p ?
       fels := Filtered(els,x -> not x in pg);  # and here?? A fraction perhaps...
# find representative involutions for the orbits of stb
#       cr := [];
#       while (Size(fels) > 0) do
#         Append(cr,[fels[1]]);
#         fels := Filtered(fels, x -> not x in Orbit(stb,fels[1]));
#         fels := Filtered(fels, x -> not IsConjugate(stb,x,fels[1]));    
#                     # is there a gap function for what I am attempting here??
#       od;
       cr := List(Orbits(stb,fels),Representative);
# find those which give string diagrams
       ccr := [];
       for c in cr do
         ok := true;
         for i in [1..n-2] do
           if (Order(c*p[i]) > 2) then    # could invalid elements be stored with each involution ?
              ok := false;
              break;
           fi;
         od;
         if IsInt(type[n-1]) then
            if not (Order(c*p[n-1])=type[n-1]) then
               ok := false;
            fi;
         else
            if not (Order(c*p[n-1]) in type[n-1]) then
               ok := false;
            fi;
         fi;
         if ok then
           Append(ccr,[c]);
         fi;
       od;
# found some.
       Append(ppp,List(ccr,x -> [Concatenation(p,[x]),Stabilizer(stb,x)]));
#     fi;     
   od;
   Print("## checking intersection property on the ",Size(ppp)," found.\n");
# check intersection property
   rtn := [[],fp];
   for stp in ppp do
     pp := stp[1];
     ok := true;
     for i in [1..n-2] do
       for j in [i+1..n-1] do
         for k in [j-1..n-1] do
           if ok then
             hn := Group(List([i..k], z -> pp[z]));
             h0 := Group(List([j..n], z -> pp[z]));
             h0n := Group([()]);
             if (k >= j) then
               h0n := Group(List([j..k], z -> pp[z]));
             fi;
             if not (Size(h0n) = Size(Intersection(h0,hn))) then
               ok := false;
             fi;
           fi;
         od;
       od;
     od;
     if (ok) then 
       if Size(Group(stp[1])) = Size(grp) then
         Append(rtn[2],[stp]);
       else 
         Append(rtn[1],[stp]);
       fi;
     fi;
   od;
   return rtn;
end;              
   

SomePolytopes := function(grp,type)
  local sz,n;
  Print("## Group of size ",Size(grp),". Now calculating Aut(G)\n");
  Print("## Automorphism Group of size ",Size(AutomorphismGroup(grp)),"\n");
  n := Length(type)+1;
  return List(SomePolytopes1(grp,n,[],type)[2], x -> x[1]);
end;

BPSL := function(q) 
  local L,n,p,g;
  L := PSL(2,q); 
  n := Size(MovedPoints(L));
  p := Product(List([1..n], i -> (i,i+n)));
  g := List(GeneratorsOfGroup(L));
  Append(g,List(g,h -> h^p));
  Append(g,[p]);
  return Group(g);
end;

l2q := [ 2, 3, 4, 
#  5,    -> removed, because L2(4) = L2(5).
  7, 8, 9, 11, 13, 16, 17, 19, 23, 25, 27, 29, 31, 32, 37, 41,
  43, 47, 49, 53, 59, 61, 64, 67, 71, 73, 79, 81, 83, 89, 97, 101, 103, 107,
  109, 113, 121, 125, 127, 128, 131, 137, 139, 149, 151, 157, 163, 167, 169,
  173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 243,
  251, 256, 257, 263, 269, 271, 277, 281, 283, 289, 293, 307, 311, 313, 317,
  331, 337, 343, 347, 349, 353, 359, 361, 367, 373, 379, 383, 389, 397, 401,
  409, 419, 421, 431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491,
  499, 503, 509, 512, 521, 523, 529, 541, 547, 557, 563, 569, 571, 577, 587,
  593, 599, 601, 607, 613, 617, 619, 625, 631, 641, 643, 647, 653, 659, 661,
  673, 677, 683, 691, 701, 709, 719, 727, 729, 733, 739, 743, 751, 757, 761,
  769, 773, 787, 797, 809, 811, 821, 823, 827, 829, 839, 841, 853, 857, 859,
  863, 877, 881, 883, 887, 907, 911, 919, 929, 937, 941, 947, 953, 961, 967,
  971, 977, 983, 991, 997 ];;

s2q := [ 2, 3, 4, 5, 7, 8, 9, 11, 13, 16, 17, 19, 23, 25, 27, 29, 31, 32, 37, 41,
  43, 47, 49, 53, 59, 61, 64, 67, 71, 73, 79, 81, 83, 89, 97, 101, 103, 107,
  109, 113, 121, 125, 127, 128, 131, 137, 139, 149, 151, 157, 163, 167, 169,
  173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 243,
  251, 256, 257, 263, 269, 271, 277, 281, 283, 289, 293, 307, 311, 313, 317,
  331, 337, 343, 347, 349, 353, 359, 361, 367, 373, 379, 383, 389, 397, 401,
  409, 419, 421, 431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491,
  499, 503, 509, 512, 521, 523, 529, 541, 547, 557, 563, 569, 571, 577, 587,
  593, 599, 601, 607, 613, 617, 619, 625, 631, 641, 643, 647, 653, 659, 661,
  673, 677, 683, 691, 701, 709, 719, 727, 729, 733, 739, 743, 751, 757, 761,
  769, 773, 787, 797, 809, 811, 821, 823, 827, 829, 839, 841, 853, 857, 859,
  863, 877, 881, 883, 887, 907, 911, 919, 929, 937, 941, 947, 953, 961, 967,
  971, 977, 983, 991, 997 ];;


AllPolytopesOfSize := function(n)
  local rtn,i,grp,num,pres,np,np2,gp,hom,img,h,tmp;
  if IsOddInt(n) then return []; fi;
  if n=2 then return [[(1,2)]]; fi;
  if IsOddInt(n/2) then 
    return [[Product(List([1..(n/2-1)/2],i -> (2*i,2*i+1))),Product(List([1..(n/2-1)/2],i ->(2*i-1,2*i)))]];
  fi;
  rtn := [];
  num := NumberSmallGroups(n);
  for i in [1..num] do
    Print("#---- Now checking SmallGroup(",n,",",i,") (out of ",num,") ----\n");
    Print("#---- ",RankList(rtn)," polytopes found  ----\n");
    grp := SmallGroup(n,i);
    hom := [IsomorphismPermGroup(grp)];
    img := [Image(hom[1])];
#    pres := PresentationViaCosetTable(grp);
#    TzGoGo(pres);
#    grp := Image(IsomorphismPermGroup(FpGroupPresentation(pres)));
    Append(hom,[SmallerDegreePermutationRepresentation(img[Length(img)])]);
    Append(img,[Image(hom[Length(hom)])]);
    while (NrMovedPoints(img[Length(img)]) < NrMovedPoints(img[Length(img)-1])) do
      Append(hom,[SmallerDegreePermutationRepresentation(img[Length(img)])]);
      Append(img,[Image(hom[Length(hom)])]);
    od;
    tmp := AllPolytopes(grp);
    for h in hom do
      tmp := List(tmp,p -> List(p,j -> j^h));
    od;
    Append(rtn,tmp);
  od;
  return rtn;
end;

SomePolytopesOfSize := function(o)
  local rtn,i,grp,num,pres,np,np2,gp,hom,img,h,tmp,n,l;
  n := o[1];
  l := o[2];
  rtn := [];
  num := NumberSmallGroups(n);
  l := Filtered(l,x -> x <= num and x > 0);
  for i in l do
    Print("#---- ",RankList(rtn)," polytopes found so far ----\n");
    Print("#--- Checking SmlGrp(",n,",",i,") (",i-Minimum(l)+1," of ",Size(l),"), NrSmlGrps = ",NrSmallGroups(n)," ----\n");
    grp := SmallGroup(n,i);
    hom := [IsomorphismPermGroup(grp)];
    img := [Image(hom[1])];
#    pres := PresentationViaCosetTable(grp);
#    TzGoGo(pres);
#    grp := Image(IsomorphismPermGroup(FpGroupPresentation(pres)));
    Append(hom,[SmallerDegreePermutationRepresentation(img[Length(img)])]);
    Append(img,[Image(hom[Length(hom)])]);
    while (NrMovedPoints(img[Length(img)]) < NrMovedPoints(img[Length(img)-1])) do
      Append(hom,[SmallerDegreePermutationRepresentation(img[Length(img)])]);
      Append(img,[Image(hom[Length(hom)])]);
    od;
    tmp := AllPolytopes(grp);
    for h in hom do
      tmp := List(tmp,p -> List(p,j -> j^h));
    od;
    Append(rtn,tmp);
  od;
  return rtn;
end;


Nonuniversality := function(pp)
  local tmp,n;
  n := Size(pp);
  tmp := Group(pp);
  tmp := PresentationViaCosetTable(tmp);
  tmp := FpGroupPresentation(tmp);
  tmp := RelatorsOfFpGroup(tmp);
  tmp := Filtered(tmp, w -> Size(Set(LetterRepAssocWord(w))) = n);
  return tmp;
end;

IsConfirmedUniversal := function(pp) 
  if Size(pp) = 1 then return true; fi;
  return Size(Nonuniversality(pp)) = 0;
end;

Facet := function(pp) 
  return pp{[1..Size(pp)-1]};
end;

VertexFigure := function(pp) 
  return pp{[2..Size(pp)]};
end;

OrIndex := function(pp) 
  local gp1,gp2;
  gp1 := Group(pp);
  gp2 := Group(List([1..Size(pp)-1],i -> pp[i]*pp[i+1]));
  return Index(gp1,gp2);
end;

IsOrientable := function(pp) 
  if (Size(pp)=1) then return true; fi;
  return OrIndex(pp)=2;
end;


fVec := function(pp) 
  local rtn,sz;
  rtn := List([1..Size(pp)]);
  rtn := List(rtn, x -> Filtered(rtn, y -> not x=y));
  sz := Size(Group(pp));
  if (sz = 2) then
    return [2];
  fi;
  rtn := List(rtn,x -> List(x,i -> pp[i]));
  rtn := List(rtn,x -> sz/Size(Group(x)));
  return rtn;
end;

IsFlat := function(pp)
  if Size(pp) < 2 then return false; fi;
  return fVec(pp)[1] = fVec(Facet(pp))[1];
end;

Petrial := function(pp)
  local rtn;
  if not Size(pp) = 3 then return fail; fi;
  rtn := List(pp);
  rtn[1] := rtn[1]*rtn[3];
  if not IsPolytope(rtn) then return fail; fi;
  return rtn;
end;

Facetting := function(pp,k)
  local rtn;
  if not Size(pp) = 3 then return fail; fi;
  rtn := List(pp);
  rtn[2] := rtn[2]*(rtn[3]*rtn[2])^(k-1);
  if not IsPolytope(rtn) then return fail; fi;
  return rtn;
end;

Halving := function(pp)
  local rtn;
  if not Size(pp) = 3 then return fail; fi;
  if not (Order(pp[1]*pp[2]) = 4) then return fail; fi;
  rtn := [pp[1]*pp[2]*pp[1],pp[3],pp[2]];
  if not IsPolytope(rtn) then return fail; fi;
  return rtn;
end;

Skewing := function(pp)
  local rtn;
  if not Size(pp) = 3 then return fail; fi;
  if not (Order(pp[3]*pp[2]) = 4) then return fail; fi;
  rtn := [pp[2],pp[1]*pp[3],(pp[2]*pp[3])^2];
  if not IsPolytope(rtn) then return fail; fi;
  return rtn;
end;

Composition := function(p1,p2)
  local g1,g2,dp,e1,e2,rtn;
  g1 := Group(p1);
  g2 := Group(p2);
  dp := DirectProduct(g1,g2);
  e1 := Embedding(dp,1);
  e2 := Embedding(dp,2);
  rtn := List(p1,x -> Image(e1,x));
  Append(rtn,List(p2,x -> Image(e2,x)));
  return rtn;
end;

IsSphericalType := function(pp) 
  local pt;
  if Size(pp)<3 then return true; fi;
  pt := PolytopeType(pp);
  if Set(pt)=[3] then return true; fi;
  if Maximum(pt) > 5 then return false; fi;
  if Size(pp)=3 then
    return 3 in pt;
  fi;
  if Size(pp)=4 then
    if pt[2]=5 then return false; fi;
    if pt[2]=4 then return pt[1]=3 and pt[3]=3; fi;
    return 3 in [pt[1],pt[3]];
  fi;
  if Maximum(pt)=5 then return false; fi;
  return Set(PolytopeType(Facet(pp)))=[3] or Set(PolytopeType(VertexFigure(pp)))=[3];
end;  

IsSpherical := function(pp) 
  local pt,sz,n;
  if not(IsSphericalType(pp)) then return false; fi;
  if Size(pp) < 3 then return true; fi;
  pt := PolytopeType(pp);
  if Set(pt)=[3] then return true; fi;
  sz := Size(Group(pp));
  if pt = [3,4,3] then return sz=1152; fi;
  if pt = [3,3,5] then return sz=14400; fi;
  if pt = [5,3,3] then return sz=14400; fi;
  if pt = [3,5] then return sz=120; fi;
  if pt = [5,3] then return sz=120; fi;
  n := Size(pp);
  return sz = 2^n*Factorial(n);
end;

IsProjective := pp -> IsSphericalType(pp) and not IsSpherical(pp);

IsLocallySpherical := pp -> Size(pp)>1 and IsSpherical(Facet(pp)) and IsSpherical(VertexFigure(pp));

IsLocallySphericalType := pp -> Size(pp)>1 and IsSphericalType(Facet(pp)) and IsSphericalType(VertexFigure(pp));

IsLocallyProjective := pp -> IsLocallySphericalType(pp) and not IsLocallySpherical(pp);

IsEuclideanType := function(pp) 
  local pt,n;
# assumes pp is finite, so {\infty} is excluded
  if Size(pp)<3 then return false; fi;
  pt := PolytopeType(pp);
  if pt=[3,6] then return true; fi;
  if pt=[6,3] then return true; fi;
  if pt=[4,4] then return true; fi;
  if Size(pp)=3 then return false; fi;
  if pt=[3,4,3,3] then return true; fi;
  if pt=[3,3,4,3] then return true; fi;
  n := Size(pt);
  if pt[1]=4 and pt[n]=4 and Maximum(pt{[2..n-1]})=3 then
    return true;
  fi;
  return false;
end;

IsToroidal := function(pp)
# assumes pp is finite, and uses results of article by McM, Sch and me.
  if Size(pp)<3 then return false; fi;
#  if Size(pp)=2 then return true; fi;
# n-gons, while tori, are more naturally regarded as circles (ie spheres)
  return IsEuclideanType(pp) and IsLocallySpherical(pp);
end;

IsLocallyToroidal := function(pp) 
  if Size(pp) < 3 then return false; fi;
  if IsLocallySpherical(pp) then return false; fi;
  if IsSpherical(Facet(pp)) or IsToroidal(Facet(pp)) then
    if IsSpherical(VertexFigure(pp)) or IsToroidal(VertexFigure(pp)) then
      return true;
    fi;
  fi;
  return false;
end;  

IsHyperbolicType := function(pp) 
# uses Thm 3C1 of Mc&Sc's book.
  if Size(pp)<3 then return false; fi;
  if IsSphericalType(pp) or IsEuclideanType(pp) then return false; fi;
  if IsSphericalType(Facet(pp)) or IsEuclideanType(Facet(pp)) then
    if  IsSphericalType(VertexFigure(pp)) or IsEuclideanType(VertexFigure(pp)) then
      return true;
    fi;
  fi;
  return false;
end;


IsCompactHyperbolicType := function(pp)
# uses Thm 3C1 of Mc&Sc's book.
  if Size(pp)<3 then return false; fi;
  if IsSphericalType(pp) or IsEuclideanType(pp) then return false; fi;
  return IsLocallySphericalType(pp);
end;

# assumes pp is finite
IsHyperbolicQuotient := pp -> Size(pp)>2 and IsHyperbolicType(pp) and IsLocallySpherical(pp);

# assumes pp is finite
IsCompactHyperbolicQuotient := pp -> Size(pp)>2 and IsCompactHyperbolicType(pp) and IsLocallySpherical(pp);

IsDegenerate := function(pp) 
  return 2 in PolytopeType(pp);
end;


AmalgamationFp := function(p1,p2)
  local tmp1,tmp2,n1,n2,fg,fgg;
  n1 := Size(p1);
  n2 := Size(p2);
  if not n1 = n2 then return fail; fi;
#  if not IsIsom(VertexFigure(p1),Facet(p2)) then return fail; fi;
  tmp1 := Group(p1);
  tmp1 := PresentationViaCosetTable(tmp1);
  tmp1 := FpGroupPresentation(tmp1);
  tmp1 := RelatorsOfFpGroup(tmp1);
  tmp1 := List(tmp1,TietzeWordAbstractWord);
  tmp2 := Group(p2);
  tmp2 := PresentationViaCosetTable(tmp2);
  tmp2 := FpGroupPresentation(tmp2);
  tmp2 := RelatorsOfFpGroup(tmp2);
  tmp2 := List(tmp2,TietzeWordAbstractWord);
  tmp2 := List(tmp2,x -> List(x,i -> i+1));
  Add(tmp1,[1,n1+1,1,n1+1]);
  tmp1 := Union(tmp1,tmp2);
  fg := FreeGroup(n1+1);
  fgg := GeneratorsOfGroup(fg);
  tmp1 := List(tmp1,x -> AbstractWordTietzeWord(x,fgg));
  return fg/tmp1;
end;

SelfDualAmalgamationFp := function(p1)
  local tmp1,tmp2,n,fg,fgg;
  n := Size(p1);
  if not n1 = n2 then return fail; fi;
#  if not IsIsom(VertexFigure(p1),Facet(p2)) then return fail; fi;
  tmp1 := Group(p1);
  tmp1 := PresentationViaCosetTable(tmp1);
  tmp1 := FpGroupPresentation(tmp1);
  tmp1 := RelatorsOfFpGroup(tmp1);
  tmp1 := List(tmp1,TietzeWordAbstractWord);
  tmp2 := List(tmp1,x -> List(x,i -> n+2-i));
  Add(tmp1,[1,n+1,1,n+1]);
  tmp1 := Union(tmp1,tmp2);
  fg := FreeGroup(n+1);
  fgg := GeneratorsOfGroup(fg);
  tmp1 := List(tmp1,x -> AbstractWordTietzeWord(x,fgg));
  return fg/tmp1;
end;


